from shexer.io.sparql.query import \
    query_endpoint_po_of_an_s, \
    query_endpoint_single_variable, \
    query_endpoint_sp_of_an_o
from shexer.model.graph.abstract_sgraph import SGraph
from shexer.model.graph.rdflib_sgraph import RdflibSgraph
from shexer.utils.uri import remove_corners
from rdflib import Graph
from shexer.consts import RDF_TYPE

_DEF_SUBJ_VARIABLE = "?s"
_DEF_SUBJ_ID = "s"

_DEF_PRED_VARIABLE = "?p"
_DEF_PRED_ID = "p"

_DEF_OBJ_VARIABLE = "?o"
_DEF_OBJ_ID = "o"

class EndpointSGraph(SGraph):

    def __init__(self, endpoint_url, store_locally=True):
        super().__init__()
        self._endpoint_url = endpoint_url
        self._store_locally = store_locally
        self._local_sgraph = RdflibSgraph(rdflib_graph=Graph()) if store_locally else None
        self._subjects_tracked = set() if store_locally else None
        self._objects_tracked = set() if store_locally else None



    def query_single_variable(self, str_query, variable_id):
        return query_endpoint_single_variable(variable_id=variable_id,
                                              str_query=str_query,
                                              endpoint_url=self._endpoint_url)

    def serialize_current_local_sgraph(self, path_file, format):
        if self._local_sgraph is not None:
            self._local_sgraph.serialize(path=path_file,
                                         format=format)


    def yield_class_triples_of_an_s(self, target_node, instantiation_property):
        if not self._store_locally:
            for a_triple in self._yield_remote_class_triples_of_an_s(target_node, instantiation_property):
                yield a_triple
        else:
            for a_triple in self._yield_local_class_triples_of_an_s(target_node, instantiation_property):
                yield a_triple

    def yield_classes_with_instances(self, instantiation_property=RDF_TYPE):
        str_query = "SELECT distinct {0} where {{ {1} <{2}> {0} . }}".format(_DEF_OBJ_VARIABLE,
                                                                             _DEF_SUBJ_VARIABLE,
                                                                              remove_corners(
                                                                                  a_uri=instantiation_property,
                                                                                  raise_error_if_no_corners=False)
                                                                              )
        for an_elem in query_endpoint_single_variable(endpoint_url=self._endpoint_url,
                                                      str_query=str_query,
                                                      variable_id=_DEF_OBJ_ID):
            yield str(an_elem)


    def _yield_remote_class_triples_of_an_s(self, target_node, instantiation_property):
        str_query = "SELECT {0} WHERE {{ <{1}> <{2}> {0} . }}".format(_DEF_OBJ_VARIABLE,
                                                                      remove_corners(a_uri=target_node,
                                                                                     raise_error_if_no_corners=False),
                                                                      remove_corners(a_uri=instantiation_property,
                                                                                     raise_error_if_no_corners=False))
        for an_elem in query_endpoint_single_variable(endpoint_url=self._endpoint_url,
                                                      str_query=str_query,
                                                      variable_id=_DEF_OBJ_ID):
            yield ("<" + target_node + ">", "<" + instantiation_property + ">", an_elem)


    def _yield_local_class_triples_of_an_s(self, target_node, instantiation_property):
        if target_node not in self._subjects_tracked:
            for a_triple in self._yield_remote_class_triples_of_an_s(target_node, instantiation_property):
                self._store_triple_locally(a_triple)
            self._subjects_tracked.add(target_node)
        for a_triple in self._local_sgraph.yield_class_triples_of_an_s(target_node, instantiation_property):
            yield a_triple


    def yield_p_o_triples_of_an_s(self, target_node):
        if not self._store_locally:
            for a_triple in self._yield_remote_p_o_triples_of_an_s(target_node):
                yield a_triple
        else:
            for a_triple in self._yield_local_p_o_triples_of_an_s(target_node):
                yield a_triple

    def yield_s_p_triples_of_an_o(self, target_node):
        if not self._store_locally:
            for a_triple in self._yield_remote_s_p_triples_of_an_o(target_node):
                yield a_triple
        else:
            for a_triple in self._yield_local_s_p_triples_of_an_o(target_node):
                yield a_triple


    def _yield_remote_p_o_triples_of_an_s(self, target_node):
        str_query = "SELECT {0} {1} WHERE {{ <{2}> {0} {1} .}} ".format(_DEF_PRED_VARIABLE,
                                                                        _DEF_OBJ_VARIABLE,
                                                                        remove_corners(a_uri=target_node,
                                                                                       raise_error_if_no_corners=False))
        for a_tuple_po in query_endpoint_po_of_an_s(endpoint_url=self._endpoint_url,
                                                    str_query=str_query,
                                                    p_id=_DEF_PRED_ID,
                                                    o_id=_DEF_OBJ_ID):
            yield "<" + target_node + ">", a_tuple_po[0], a_tuple_po[1]

    def _yield_remote_s_p_triples_of_an_o(self, target_node):
        str_query = "SELECT {0} {1} WHERE {{ {0} {1} <{2}> .}}".format(_DEF_SUBJ_VARIABLE,
                                                                       _DEF_PRED_VARIABLE,
                                                                       remove_corners(a_uri=target_node,
                                                                                      raise_error_if_no_corners=False))
        for a_tuple_sp in query_endpoint_sp_of_an_o(endpoint_url=self._endpoint_url,
                                                    str_query=str_query,
                                                    p_id=_DEF_PRED_ID,
                                                    s_id=_DEF_SUBJ_ID):
            yield a_tuple_sp[0], a_tuple_sp[1], "<" + target_node + ">"


    def _yield_local_p_o_triples_of_an_s(self, target_node):
        if target_node not in self._subjects_tracked:
            for a_triple in self._yield_remote_p_o_triples_of_an_s(target_node):
                self._store_triple_locally(a_triple)
            self._subjects_tracked.add(target_node)
        for a_triple in self._local_sgraph.yield_p_o_triples_of_an_s(target_node):
            yield a_triple

    def _yield_local_s_p_triples_of_an_o(self, target_node):
        if target_node not in self._objects_tracked:
            for a_triple in self._yield_remote_s_p_triples_of_an_o(target_node):
                self._store_triple_locally(a_triple)
            self._objects_tracked.add(target_node)
        for a_triple in self._local_sgraph.yield_s_p_triples_of_an_o(target_node):
            yield a_triple

    def _store_triple_locally(self, a_triple):
        self._local_sgraph.add_triple(a_triple)


from shexer.consts import NT, FIXED_SHAPE_MAP, SHAPES_DEFAULT_NAMESPACE
from shexer.utils.factories.triple_yielders_factory import get_triple_yielder, tune_target_classes_if_needed, \
    read_target_classes_from_file
from shexer.core.instances.instance_tracker import InstanceTracker
from shexer.core.instances.mappings.shape_map_instance_tracker import ShapeMapInstanceTracker
from shexer.core.instances.mix.mixed_instance_tracker import MixedInstanceTracker
from shexer.utils.factories.iri_factory import create_IRIs_from_string_list
from shexer.utils.factories.shape_map_parser_factory import get_shape_map_parser
from shexer.model.graph.endpoint_sgraph import EndpointSGraph
from shexer.model.graph.rdflib_sgraph import RdflibSgraph
from shexer.utils.dict import reverse_keys_and_values


def get_instance_tracker(instances_file_input=None, graph_file_input=None,
                         graph_list_of_files_input=None, target_classes=None,
                         file_target_classes=None, input_format=NT,
                         instantiation_property=None,
                         infer_numeric_types_for_untyped_literals=None,
                         namespaces_to_ignore=None,
                         raw_graph=None,
                         all_classes_mode=False,
                         namespaces_dict=None,
                         url_input=None,
                         list_of_url_input=None,
                         rdflib_graph=None,
                         shape_map_file=None,
                         shape_map_raw=None,
                         shape_map_format=FIXED_SHAPE_MAP,
                         track_classes_for_entities_at_last_depth_level=True,
                         depth_for_building_subgraph=1,
                         url_endpoint=None,
                         strict_syntax_with_corners=False,
                         namespaces_for_qualifier_props=None,
                         shape_qualifiers_mode=False,
                         built_remote_graph=None,
                         built_shape_map=None,
                         shapes_namespace=SHAPES_DEFAULT_NAMESPACE,
                         limit_remote_instances=-1,
                         inverse_paths=False,
                         compression_mode=None,
                         disable_endpoint_cache=False,
                         instances_cap=-1
                         ):
    """

    :param instances_file_input:
    :param graph_file_input:
    :param graph_list_of_files_input:
    :param target_classes:
    :param file_target_classes:
    :param input_format:
    :param instantiation_property:
    :param namespaces_to_ignore:
    :param raw_graph:
    :param all_classes_mode:
    :param namespaces_dict:
    :param url_input:
    :param list_of_url_input:
    :param shape_map_file:
    :param shape_map_raw:
    :param shape_map_format:
    :param track_classes_for_entities_at_last_depth_level:
    :param depth_for_building_subgraph:
    :param url_endpoint:
    :param strict_syntax_with_corners:
    :param namespaces_for_qualifier_props:
    :param shape_qualifiers_mode:
    :param built_remote_graph:
    :param built_shape_map:
    :return:
    """

    prefix_namespaces_dict = reverse_keys_and_values(namespaces_dict)
    instance_yielder = None  # Old-schooler
    if instances_file_input is not None:
        instance_yielder = get_triple_yielder(source_file=instances_file_input,
                                              input_format=input_format,
                                              namespaces_to_ignore=namespaces_to_ignore,
                                              raw_graph=raw_graph,
                                              namespaces_dict=namespaces_dict,
                                              allow_untyped_numbers=infer_numeric_types_for_untyped_literals,
                                              url_input=url_input,
                                              list_of_url_input=list_of_url_input,
                                              rdflib_graph=rdflib_graph,
                                              instantiation_property=instantiation_property,
                                              shape_map_file=shape_map_file,
                                              shape_map_raw=shape_map_raw,
                                              track_classes_for_entities_at_last_depth_level=
                                              track_classes_for_entities_at_last_depth_level,
                                              depth_for_building_subgraph=depth_for_building_subgraph,
                                              url_endpoint=url_endpoint,
                                              strict_syntax_with_corners=strict_syntax_with_corners,
                                              target_classes=target_classes,
                                              file_target_classes=file_target_classes,
                                              built_remote_graph=built_remote_graph,
                                              built_shape_map=built_shape_map,
                                              limit_remote_instances=limit_remote_instances,
                                              inverse_paths=inverse_paths,
                                              all_classes_mode=all_classes_mode,
                                              compression_mode=compression_mode,
                                              disable_endpoint_cache=disable_endpoint_cache
                                              )
    else:
        instance_yielder = get_triple_yielder(source_file=graph_file_input,
                                              list_of_source_files=graph_list_of_files_input,
                                              input_format=input_format,
                                              namespaces_to_ignore=namespaces_to_ignore,
                                              raw_graph=raw_graph,
                                              namespaces_dict=namespaces_dict,
                                              allow_untyped_numbers=infer_numeric_types_for_untyped_literals,
                                              url_input=url_input,
                                              list_of_url_input=list_of_url_input,
                                              rdflib_graph=rdflib_graph,
                                              instantiation_property=instantiation_property,
                                              shape_map_file=shape_map_file,
                                              shape_map_raw=shape_map_raw,
                                              track_classes_for_entities_at_last_depth_level=track_classes_for_entities_at_last_depth_level,
                                              depth_for_building_subgraph=depth_for_building_subgraph,
                                              url_endpoint=url_endpoint,
                                              strict_syntax_with_corners=strict_syntax_with_corners,
                                              target_classes=target_classes,
                                              file_target_classes=file_target_classes,
                                              built_remote_graph=built_remote_graph,
                                              built_shape_map=built_shape_map,
                                              limit_remote_instances=limit_remote_instances,
                                              inverse_paths=inverse_paths,
                                              all_classes_mode=all_classes_mode,
                                              compression_mode=compression_mode,
                                              disable_endpoint_cache=disable_endpoint_cache
                                              )

    selectors_tracker = None
    pure_instances_tracker = None

    if _are_there_selectors(shape_map_file, shape_map_raw):
        sgraph = _get_adequate_sgraph(endpoint_url=url_endpoint,
                                      raw_graph=raw_graph,
                                      graph_file_input=graph_file_input,
                                      url_input=url_input,
                                      graph_format=input_format,
                                      built_remote_graph=built_remote_graph,
                                      disable_endpoint_cache=disable_endpoint_cache)
        valid_shape_map = built_shape_map
        if built_shape_map is None:
            shape_map_parser = get_shape_map_parser(format=shape_map_format,
                                                    sgraph=sgraph,
                                                    namespaces_prefix_dict=namespaces_dict)
            valid_shape_map = shape_map_parser.parse_shape_map(source_file=shape_map_file,
                                                               raw_content=shape_map_raw)
        selectors_tracker = ShapeMapInstanceTracker(shape_map=valid_shape_map)
    if _are_there_some_target_classes(target_classes, file_target_classes, all_classes_mode, shape_qualifiers_mode):
        model_classes = None
        if file_target_classes or target_classes is not None:
            list_of_str_target_classes = tune_target_classes_if_needed(
                list_target_classes=target_classes,
                prefix_namespaces_dict=prefix_namespaces_dict) if target_classes is not None else read_target_classes_from_file(
                file_target_classes=file_target_classes,
                prefix_namespaces_dict=prefix_namespaces_dict)
            model_classes = get_list_of_model_classes(list_of_str_target_classes)

        pure_instances_tracker = InstanceTracker(target_classes=model_classes,
                                                 triples_yielder=instance_yielder,
                                                 instantiation_property=instantiation_property,
                                                 all_classes_mode=all_classes_mode,
                                                 track_hierarchies=False,
                                                 namespaces_for_qualifier_props=namespaces_for_qualifier_props,
                                                 shape_qualifiers_mode=shape_qualifiers_mode,
                                                 shapes_namespace=shapes_namespace,
                                                 instances_cap=instances_cap)

    return _decide_tracker_to_return(selectors_tracker, pure_instances_tracker)


def _get_adequate_sgraph(endpoint_url, graph_file_input, url_input, graph_format,
                         raw_graph, built_remote_graph, disable_endpoint_cache):
    if endpoint_url is not None:
        return built_remote_graph if built_remote_graph is not None else EndpointSGraph(endpoint_url=endpoint_url,
                                                                                        store_locally=not disable_endpoint_cache)
    else:
        return RdflibSgraph(source_file=graph_file_input if graph_file_input is not None else url_input,
                            raw_graph=raw_graph,
                            format=graph_format)


def _decide_tracker_to_return(selectors_tracker, pure_instances_tracker):
    if selectors_tracker is not None and pure_instances_tracker is not None:
        return MixedInstanceTracker(list_of_instance_trackers=[selectors_tracker, pure_instances_tracker])
    return selectors_tracker if selectors_tracker is not None else pure_instances_tracker


def _are_there_selectors(shape_map_file, shape_map_raw):
    if shape_map_file is None and shape_map_raw is None:
        return False
    return True


def _are_there_some_target_classes(target_classes, file_target_classes, all_classes_mode, shape_qualifiers_mode):
    if target_classes is None and file_target_classes is None and not all_classes_mode and not shape_qualifiers_mode:
        return False
    return True


def get_list_of_model_classes(list_of_str_target_classes):
    return create_IRIs_from_string_list(list_of_str_target_classes)



from shexer.io.shex.formater.consts import SPACES_GAP_BETWEEN_TOKENS, \
    COMMENT_INI, TARGET_LINE_LENGHT, SPACES_GAP_FOR_FREQUENCY, KLEENE_CLOSURE, POSITIVE_CLOSURE, OPT_CARDINALITY, SHAPE_LINK_CHAR
from shexer.model.IRI import IRI_ELEM_TYPE
from shexer.model.shape import STARTING_CHAR_FOR_SHAPE_NAME
from shexer.utils.shapes import prefixize_shape_name_if_possible

_INVERSE_SENSE_SHEXC = "^"

class BaseStatementSerializer(object):

    def __init__(self, instantiation_property_str, frequency_serializer, disable_comments=False, is_inverse=False):
        self._instantiation_property_str = instantiation_property_str
        self._disable_comments = disable_comments
        self._is_inverse = is_inverse
        self._frequency_serializer = frequency_serializer

    def serialize_statement_with_indent_level(self, a_statement, is_last_statement_of_shape, namespaces_dict):
        tuples_line_indent = []
        st_property = BaseStatementSerializer.tune_token(a_statement.st_property, namespaces_dict)
        st_target_element = self.str_of_target_element(target_element=a_statement.st_type,
                                                       st_property=a_statement.st_property,
                                                       namespaces_dict=namespaces_dict)
        cardinality = BaseStatementSerializer.cardinality_representation(
            statement=a_statement,
            out_of_comment=True)
        result = self._sense_flag() + st_property + SPACES_GAP_BETWEEN_TOKENS + st_target_element + SPACES_GAP_BETWEEN_TOKENS + \
                 cardinality + BaseStatementSerializer.closure_of_statement(is_last_statement_of_shape)

        if a_statement.cardinality not in [KLEENE_CLOSURE, OPT_CARDINALITY] and not self._disable_comments:
            result += BaseStatementSerializer.adequate_amount_of_final_spaces(result)
            result += a_statement.probability_representation()
        tuples_line_indent.append((result, 1))

        for a_comment in a_statement.comments:
            tuples_line_indent.append((a_comment, 4))

        return tuples_line_indent

    def str_of_target_element(self, target_element, st_property, namespaces_dict):
        """
        Special treatment for instantiation_property. We build a value set with an specific URI
        :param target_element:
        :param st_property:
        :param namespaces_dict:
        :return:
        """
        if st_property == self._instantiation_property_str:
            return "[" + BaseStatementSerializer.tune_token(target_element, namespaces_dict) + "]"
        return BaseStatementSerializer.tune_token(target_element, namespaces_dict)

    @staticmethod
    def tune_token(a_token, namespaces_dict):
        # TODO:  a lot to correct here for normal behaviour
        if a_token.startswith(STARTING_CHAR_FOR_SHAPE_NAME):  # Shape
            # return STARTING_CHAR_FOR_SHAPE_NAME +":" + a_token.replace(STARTING_CHAR_FOR_SHAPE_NAME, "")
            return SHAPE_LINK_CHAR \
                   + prefixize_shape_name_if_possible(a_shape_name=a_token,
                                                      namespaces_prefix_dict=namespaces_dict)
        if a_token == IRI_ELEM_TYPE:  # iri
            return a_token
        if ":" not in a_token:
            if "<" in a_token:
                return SHAPE_LINK_CHAR + a_token
            else:
                return SHAPE_LINK_CHAR + "<" + a_token + ">"
        candidate_prefixed = BaseStatementSerializer._prefixize_uri_if_possible(uri=a_token,
                                                                                namespaces_dict=namespaces_dict)
        if candidate_prefixed is not None:
            return candidate_prefixed

        return "<" + a_token + ">"  # Complete URIs

    @staticmethod
    def _prefixize_uri_if_possible(uri, namespaces_dict):
        """
        It returns None it it doesnt find an adequate prefix

        :param uri:
        :param namespaces_dict:
        :return:
        """
        best_match = None
        for a_namespace in namespaces_dict:  # Prefixed element (all literals are prefixed elements)
            if uri.startswith(a_namespace):
                if "/" not in uri[len(a_namespace):] and \
                        "#" not in uri[len(a_namespace):]:
                    best_match = a_namespace
                    break

        return None if best_match is None else uri.replace(best_match, namespaces_dict[best_match] + ":")


    def probability_representation(self, statement):
        return COMMENT_INI + self._frequency_serializer.serialize_frequency(statement)

    @staticmethod
    def cardinality_representation(statement, out_of_comment=False):
        cardinality = statement.cardinality
        if out_of_comment and cardinality == 1:
            return ""
        if cardinality in [POSITIVE_CLOSURE, KLEENE_CLOSURE, OPT_CARDINALITY]:
            return cardinality
        else:
            return "{" + str(cardinality) + "}"

    @staticmethod
    def closure_of_statement(is_last_statement):
        if is_last_statement:
            return ""
        return ";"

    @staticmethod
    def adequate_amount_of_final_spaces(current_line):
        if len(current_line) > TARGET_LINE_LENGHT - 10:
            return SPACES_GAP_FOR_FREQUENCY
        result = ""
        for i in range(0, TARGET_LINE_LENGHT - len(current_line)):
            result += " "
        return result

    @staticmethod
    def turn_statement_into_comment(statement, namespaces_dict):
        return statement.probability_representation() + \
               " obj: " + BaseStatementSerializer.tune_token(statement.st_type,
                                                             namespaces_dict) + \
               ". Cardinality: " + statement.cardinality_representation()

    def _sense_flag(self):
        return "" if not self._is_inverse else _INVERSE_SENSE_SHEXC + SPACES_GAP_BETWEEN_TOKENS



from shexer.consts import SHEXC, SHACL_TURTLE
from shexer.io.shex.formater.shex_serializer import ShexSerializer
from shexer.io.shacl.formater.shacl_serializer import ShaclSerializer
from shexer.io.uml.uml_serializer import UMLSerializer
from shexer.consts import RATIO_INSTANCES, UML_PLANT_SERVER


def get_shape_serializer(output_format, shapes_list, target_file=None, string_return=False, namespaces_dict=None,
                         instantiation_property=None, disable_comments=False, wikidata_annotation=False,
                         instances_report_mode=RATIO_INSTANCES, detect_minimal_iri=False, shape_features_examples=None,
                         examples_mode=None, inverse_paths=False):
    if output_format == SHEXC:
        return ShexSerializer(target_file=target_file,
                              shapes_list=shapes_list,
                              namespaces_dict=namespaces_dict,
                              string_return=string_return,
                              instantiation_property_str=instantiation_property,
                              disable_comments=disable_comments,
                              wikidata_annotation=wikidata_annotation,
                              instances_report_mode=instances_report_mode,
                              detect_minimal_iri=detect_minimal_iri,
                              shape_example_features=shape_features_examples,
                              examples_mode=examples_mode,
                              inverse_paths=inverse_paths)
    elif output_format == SHACL_TURTLE:
        return ShaclSerializer(target_file=target_file,
                               shapes_list=shapes_list,
                               namespaces_dict=namespaces_dict,
                               string_return=string_return,
                               instantiation_property_str=instantiation_property,
                               wikidata_annotation=wikidata_annotation,
                               shape_example_features=shape_features_examples,
                               detect_minimal_iri=detect_minimal_iri)
    else:
        raise ValueError("Currently unsupported format in 'output_format': " + output_format)


def get_uml_serializer(shapes_list, image_path, url_server=UML_PLANT_SERVER, namespaces_dict=None,):
    return UMLSerializer(shapes_list=shapes_list,
                         url_server=url_server,
                         image_path=image_path,
                         namespaces_dict=namespaces_dict)


from shexer.io.graph.yielder.base_triples_yielder import BaseTriplesYielder
from shexer.utils.uri import remove_corners, unprefixize_uri_mandatory
from shexer.utils.triple_yielders import tune_subj, tune_prop, tune_token
import re

_OTHER_BLANKS = re.compile("[\r\n\t]")
_SEVERAL_BLANKS = re.compile("  +")
_QUOTES_FOR_LITERALS = re.compile('[^\\\]"')
_INIT_INLINE_COMMENT = re.compile(" #")
_RDF_TYPE_CONTRACTED = ["a", "rdf:type"]
_RDF_TYPE_URI = "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>"
_BOOLEANS = ["true", "false"]
_INI_BASE_URIS = ["/", "#"]
_CLOSURES = [",", ";", "."]
_S = 0
_P = 1
_O = 2

_WAITING_FOR_SUBJ = 0
_WAITING_FOR_PRED = 1
_WAITING_FOR_OBJ = 2
_NOT_WAITING = 4

"""
TTL parser that yield triples (model objects) without loading the whole graph content in 
main memory.

WARNING: This parser works with some frequent structural assumptions of turtle files that
are not part of the standard. You may get unexpected errors or unexpected results dealing
with files containing lines which represent more than one triple. Also, we assume a totally
wel--formated input. Bad-formatted may remain undetected and produce wrong triples.

Please, in case you do not need to parse huge files that do not fit in the main memory
of your computer, use RdflifTriplesYielder instead
"""


class BigTtlTriplesYielder(BaseTriplesYielder):

    def __init__(self, source_file=None, allow_untyped_numbers=True, raw_graph=None,
                 compression_mode=None, zip_base_archive=None):

        super(BigTtlTriplesYielder, self).__init__()
        self._source_file = source_file
        self._raw_graph = raw_graph
        self._triples_count = 0
        self._error_triples = 0
        self._allow_untyped_numbers = allow_untyped_numbers
        self._compression_mode = compression_mode
        self._line_reader = self._decide_line_reader(source_file=source_file,
                                                     raw_graph=raw_graph,
                                                     compression_mode=compression_mode,
                                                     zip_base_archive=zip_base_archive)
        # Support
        self._prefixes = {}
        self._base = None

        # To be used while parsing
        self._state = _WAITING_FOR_SUBJ
        self._tmp_s = None
        self._tmp_p = None
        self._tmp_o = None
        self._last_triple_jump = None

        self._triple_ready = False

    def yield_triples(self):
        self._reset_parsing()
        for a_line in self._line_reader.read_lines():
            for a_triple in self._process_line_2(a_line):
                self._triples_count += 1
                yield (
                    tune_subj(a_triple[_S],
                              raise_error_if_no_corners=False),
                    tune_prop(a_triple[_P],
                              raise_error_if_no_corners=False),
                    tune_token(a_triple[_O],
                               base_namespace=self._base,
                               allow_untyped_numbers=self._allow_untyped_numbers,
                               raise_error_if_no_corners=False)
                )
            # self._process_line(a_line)
            # if self._triple_ready:
            #     self._triples_count += 1
            #     yield (
            #         tune_subj(self._tmp_s,
            #                   raise_error_if_no_corners=False),
            #         tune_prop(self._tmp_p),
            #         tune_token(self._tmp_o,
            #                    base_namespace=self._base,
            #                    allow_untyped_numbers=self._allow_untyped_numbers,
            #                    raise_error_if_no_corners=False)
            #     )
            #     self._triple_ready = False


    def _clean_line(self, str_line):
        result = _OTHER_BLANKS.sub(" ", str_line)
        result = _SEVERAL_BLANKS.sub(" ", result)
        result = result.strip()
        return result if " #" not in result else self._remove_comments_if_needed(result)

    def _remove_comments_if_needed(self, str_line):
        """Remove comments in the middle of the line.
        Lines starting with # wont be erased
        """
        if '"' not in str_line:  # Comment mark and no literals, trivial case
            return str_line[:str_line.find(" #")]
        # We need to find the begining and end of the literal to avoid erasing
        # comments within literals (actual content)
        quotes_indexes = []
        count_down_quotes = 2
        for a_match in _QUOTES_FOR_LITERALS.finditer(str_line):
            quotes_indexes.append(a_match.start(0))
            count_down_quotes -= 1
            if count_down_quotes == 0:
                break
        for a_match in _INIT_INLINE_COMMENT.finditer(str_line):
            if a_match.start(0) < quotes_indexes[0] or a_match.start(0) > quotes_indexes[1]:
                return str_line[:a_match.start(0)]
        return str_line  # If this point is reached, it means that the potential comments
                         # are actual content of a string literal

    def _process_line_2(self, str_line):
        str_line = self._clean_line(str_line)
        if str_line == "":
            self._process_empty_line(str_line)
        elif str_line.startswith("@prefix"):
            self._process_prefix_line(str_line)
        elif str_line.startswith("@base"):
            self._process_base_line(str_line)
        elif str_line.startswith("#"):
            self._process_comment_line(str_line)
        else:
            for a_triple in self._process_line_with_potential_triples(str_line):
                yield a_triple

    def _process_line_with_potential_triples(self, a_line):
        next_token, next_index = self._next_line_token(a_line, 0)
        while next_token != None:
            if next_token == ",":
                yield self._current_triple()
                self._state = _WAITING_FOR_OBJ
            elif next_token == ";":
                yield self._current_triple()
                self._state = _WAITING_FOR_PRED
            elif next_token == ".":
                yield self._current_triple()
                self._state = _WAITING_FOR_SUBJ
            else:
                self._assing_tmp_element_and_promote_state(next_token)
            next_token, next_index = self._next_line_token(a_line, next_index)

    def _current_triple(self):
        return self._tmp_s, self._tmp_p, self._tmp_o

    def _assing_tmp_element_and_promote_state(self, token):
        if self._state == _WAITING_FOR_SUBJ:
            self._tmp_s = self._parse_elem(token)
            self._state = _WAITING_FOR_PRED
        elif self._state == _WAITING_FOR_PRED:
            self._tmp_p = self._parse_elem(token)
            self._state = _WAITING_FOR_OBJ
        elif self._state == _WAITING_FOR_OBJ:
            self._tmp_o = self._parse_elem(token)
            self._state = _NOT_WAITING
        else:
            raise ValueError("Malformed file. Processing an unexpected token: " + token)


    def _next_line_token(self, a_line, start_index):
        while(start_index < len(a_line) and a_line[start_index] == " "):
            start_index += 1
        if start_index >= len(a_line):
            return None, None
        if a_line[start_index] in _CLOSURES:
            return a_line[start_index], start_index + 1
        elif a_line[start_index] == "<":
            end_index = a_line.find(">", start_index)
            return self._parse_cornered_element(cornered_element=a_line[start_index:end_index+1]), end_index + 1
        elif a_line[start_index] == '"':
            end_index = self._find_next_quoted_literal_ending(a_line, start_index)
            return a_line[start_index:end_index+1], end_index + 1
        else:  # could be a prefixed element, a bnode, a non-string literal... find the next blank anyway
            end_index = self._find_next_blank(a_line, start_index)
            return a_line[start_index:end_index], end_index + 1


    def _find_next_blank(self, target_str, start_index):
        pos = target_str.find(" ", start_index)
        return len(target_str)-1 if pos == -1 else pos


    def _find_next_unescaped_quotes(self, target_str, start_index):
        pos = target_str.find('"', start_index)
        while pos != -1:
            if target_str[pos-1] != "\\":
                return pos  # not escaped
            # if pos >= 2 and target_str[pos-2] == '\\':
            #     return pos  # the scape is scaped, so not escaped
            if self._count_prior_backslashes(an_str=target_str,
                                             quote_pos=pos) % 2 == 0:
                return pos # the scape is scaped, so not escaped
            pos = target_str.find('"', pos+1)
        if pos == -1:
            raise ValueError("Is this line malformed? Can`t find quotes matching: " + target_str)

    def _count_prior_backslashes(self, an_str, quote_pos):
        """
        We assume that there is at least a backslash at an_str[pos-1], so pos-1 is a non-negative index of an_str
        """
        counter = 1
        quote_pos -= 2
        while quote_pos >= 0:
            if an_str[quote_pos] == "\\":
                counter += 1
            else:
                return counter
            quote_pos -= 1
        return counter


    def _find_next_quoted_literal_ending(self, target_str, start_index):
        next_quotes = self._find_next_unescaped_quotes(target_str=target_str,
                                                       start_index=start_index+1)
        if next_quotes +1 > len(target_str) or target_str[next_quotes + 1] == " ":
            return next_quotes
        elif target_str[next_quotes + 1] == "^":
            return self._find_next_blank(target_str, next_quotes) - 1
        else:
            raise ValueError("Malformed literal? It seems like there is a problem of unmatching quotes: " + target_str)

    def _process_line(self, str_line):
        str_line = self._clean_line(str_line)
        if str_line == "":
            self._process_empty_line(str_line)
        elif '"' in str_line:
            self._process_line_with_literal(str_line)
        elif str_line.startswith("@prefix"):
            self._process_prefix_line(str_line)
        elif str_line.startswith("@base"):
            self._process_base_line(str_line)
        elif str_line.startswith("#"):
            self._process_comment_line(str_line)
        elif str_line[-1] in [",", ".", ";"]:
            if ", " in str_line[:-1]:
                # If there is a comma in a URI, it can't be followed by a blank
                self._process_multi_triple_line_commas(str_line)
            else:
                self._process_single_triple_line(str_line)
        elif " " not in str_line:
            if len(str_line) > 1:  # We are ensuring that this is not a single char, such as "," or "."
                self._process_isolated_subject(str_line)
        else:
            self._process_unknown_line(str_line)

    def _process_line_with_literal(self, line):
        first_quotes_index = line.find('"')
        s_o_line = line[:first_quotes_index].strip()
        s_o_pieces = s_o_line.split(" ")
        if len(s_o_pieces) == 2:
            self._tmp_s = self._parse_elem(s_o_pieces[0])
            self._tmp_p = self._parse_elem(s_o_pieces[1])
        elif len(s_o_pieces) == 1 and s_o_pieces[0] != "":
            self._tmp_p = self._parse_elem(s_o_pieces[0])
        # The last char MUST be in [,.;] since this lines comes stripped.
        # SO everything between first_quotes_index and line[-1], stripped
        # should be out target literal (typed or not)
        self._tmp_o = line[first_quotes_index:-1].rstrip()
        self._decide_current_triple()

    def _process_prefix_line(self, line):
        pieces = line.split(" ")
        prefix = pieces[1] if not pieces[1].endswith(":") else pieces[1][: - 1]
        base_url = remove_corners(pieces[2])
        self._prefixes[prefix] = base_url

    def _process_base_line(self, line):
        pieces = line.split(" ")
        # base_url = pieces[1] if not pieces[1].endswith(":") else pieces[1][: - 1]
        # base_url = remove_corners(pieces[2])
        self._base = remove_corners(pieces[1])

    def _process_comment_line(self, line):
        pass  # At this point, just ignore it.

    def _process_empty_line(self, line):
        pass  # At this point, just ignore it.

    def _process_unknown_line(self, line):
        self._error_triples += 1


    def _process_multi_triple_line_commas(self, line):
        pieces = line.split(" ")
        index_first_comma = 0
        for i in range(0, len(pieces)):
            if pieces[i] == ",":
                index_first_comma = i
                break
        if index_first_comma == 3:
            self._tmp_s = self._parse_elem(pieces[0])
            self._tmp_p = self._parse_elem(pieces[1])
            self._tmp_o = self._parse_elem(pieces[2])
        elif index_first_comma == 2:
            self._tmp_p = self._parse_elem(pieces[0])
            self._tmp_o = self._parse_elem(pieces[1])
        elif index_first_comma == 1:
            self._tmp_o = self._parse_elem(pieces[0])
        # else impossible?
        self._decide_current_triple()

        for i in range(index_first_comma + 2, len(pieces), 2):
            self._tmp_o = self._parse_elem(pieces[i - 1])
            self._decide_current_triple()

    def _process_single_triple_line(self, line):
        pieces = line.split(" ")
        if len(pieces) == 4:
            self._tmp_s = self._parse_elem(pieces[0])
            self._tmp_p = self._parse_elem(pieces[1])
            self._tmp_o = self._parse_elem(pieces[2])

        elif len(pieces) == 3:
            self._tmp_p = self._parse_elem(pieces[0])
            self._tmp_o = self._parse_elem(pieces[1])
        elif len(pieces) == 2:
            self._tmp_o = self._parse_elem(pieces[0])
        self._decide_current_triple()

    def _process_isolated_subject(self, line):
        # No splitt. Line is expected to contain a line with no blanks (isolated subject)
        self._tmp_s = self._parse_elem(line)
        # No need to decide triple now, incomplete element

    def _decide_current_triple(self):
        # if self._is_bnode(self._tmp_s):
        #     self._ignored_triples += 1
        # elif self._is_bnode(self._tmp_o):
        #     self._ignored_triples += 1
        # elif self._is_num_literal(self._tmp_o):
        #     self._ignored_triples += 1
        # elif self._is_boolean(self._tmp_o):
        #     self._ignored_triples += 1
        # else:
        self._triple_ready = True

    def _is_boolean(self, raw_element):
        return True if raw_element in _BOOLEANS else False

    def _is_bnode(self, a_elem):
        if a_elem[0] == "_":
            return True
        return False

    def _is_num_literal(self, elem):
        try:
            float(elem)
            return True
        except ValueError:
            return False

    def _parse_elem(self, raw_elem):
        if raw_elem[0] == "<":
            return self._parse_cornered_element(raw_elem)
        elif raw_elem in _RDF_TYPE_CONTRACTED:
            return _RDF_TYPE_URI
        elif raw_elem.startswith('"'):  # it's a literal, will be better parsed later
            return raw_elem
        elif ":" in raw_elem:
            if raw_elem.startswith("_:"):
                return raw_elem
            return unprefixize_uri_mandatory(target_uri=raw_elem,
                                             prefix_namespaces_dict=self._prefixes)
        elif raw_elem in _BOOLEANS or self._is_num_literal(raw_elem):
            return raw_elem
            # else?? shouldnt happen, let it break with a nullpoitner

    def _parse_cornered_element(self, cornered_element):
        if self._base is None:
            return cornered_element  # There is no base
        elif cornered_element[1] in _INI_BASE_URIS:
            return "<" + self._base + cornered_element[2:-1] + ">"
        elif not cornered_element[1:].startswith("http"):
            return "<" + self._base + cornered_element[1:-1] + ">"
        else:
            return cornered_element  # Nothing to do with base

    @property
    def yielded_triples(self):
        return self._triples_count

    @property
    def error_triples(self):
        return self._error_triples

    @property
    def ignored_triples(self):
        return self._ignored_triples

    def _reset_parsing(self):
        self._error_triples = 0
        self._triples_count = 0
        self._ignored_triples = 0
        self._state = _WAITING_FOR_SUBJ







class ShapeMap(object):

    def __init__(self, shape_map_items=None):
        self._items = shape_map_items if shape_map_items is not None else []


    def add_item(self, shape_map_item):
        self._items.append(shape_map_item)


    def yield_items(self):
        for an_item in self._items:
            yield an_item

    def get_sgraph(self):
        if len(self._items) == 0:
            return None
        return self._items[0].node_selector.sgraph  # Assuming they all have the same sgraph


class ShapeMapItem(object):

    def __init__(self, node_selector, shape_label):
        self._node_selector = node_selector
        self._shape_label = shape_label

    @property
    def node_selector(self):
        return self._node_selector

    @property
    def shape_label(self):
        return self._shape_label


from shexer.io.graph.yielder.base_triples_yielder import BaseTriplesYielder


class MultifileBaseTripleYielder(BaseTriplesYielder):

    def __init__(self, list_of_files,
                 namespaces_to_ignore=None,
                 allow_untyped_numbers=False,
                 compression_mode=None,
                 zip_base_archive=None):
        super(BaseTriplesYielder, self).__init__()
        self._list_of_files = list_of_files
        self._namespaces_to_ignore = namespaces_to_ignore
        self._allow_untyped_numbers = allow_untyped_numbers
        self._compression_mode = compression_mode
        self._zip_base_archive = zip_base_archive

        self._triples_yielded_from_used_yielders = 0
        self._error_triples_from_used_yielders = 0
        self._last_yielder = None


    def yield_triples(self, parse_namespaces=True):
        self._reset_count()
        for a_source_file in self._list_of_files:
            for a_triple in self._yield_triples_of_file(a_source_file, parse_namespaces):
                yield a_triple

    def _yield_triples_of_file(self, a_source_file, parse_namespaces=False):
        if self._last_yielder is not None:
            self._triples_yielded_from_used_yielders += self._last_yielder.yielded_triples
            self._error_triples_from_used_yielders += self._last_yielder.error_triples
        self._last_yielder = self._constructor_file_yielder(a_source_file=a_source_file)
        for a_triple in self._yield_triples_of_last_yielder(parse_namespaces):
            yield a_triple

    @property
    def yielded_triples(self):
        triples_current_yielder = 0 if self._last_yielder is None else self._last_yielder.yielded_triples
        return self._triples_yielded_from_used_yielders + triples_current_yielder

    @property
    def error_triples(self):
        errors_current_yielder = 0 if self._last_yielder is None else self._last_yielder.error_triples
        return self._error_triples_from_used_yielders  + errors_current_yielder

    def _reset_count(self):
        self._error_triples_from_used_yielders = 0
        self._triples_yielded_from_used_yielders = 0

    def _constructor_file_yielder(self, a_source_file):
        raise NotImplementedError("Implement in derived classes")
        
    def _yield_triples_of_last_yielder(self, parse_namespaces=True):
        """
        This is a default implementation for every yielrder (most of them) which ignores parse_namespaces
        :param parse_namespaces:
        :return:
        """
        for a_triple in self._last_yielder.yield_triples():
            yield a_triple






import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison, number_of_shapes
import os.path as pth

_BASE_DIR = BASE_FILES + "empty_shapes" + pth.sep

class TestRemoveEmptyShapes(unittest.TestCase):

    def test_one_empty_remove(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Machine"],
                        graph_file_input=G1,
                        all_classes_mode=False,
                        input_format="turtle",
                        disable_comments=True,
                        remove_empty_shapes=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(number_of_shapes(str_result) == 0)

    def test_one_empty_not_remove(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Machine"],
                        graph_file_input=G1,
                        all_classes_mode=False,
                        input_format="turtle",
                        disable_comments=True,
                        remove_empty_shapes=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "one_empty_not_remove.shex",
                                                      str_target=str_result))


    def test_some_empty_remove(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Machine",
                                        "http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        all_classes_mode=False,
                        input_format="turtle",
                        disable_comments=True,
                        remove_empty_shapes=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_some_empty_not_remove(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Machine",
                                        "http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        all_classes_mode=False,
                        input_format="turtle",
                        disable_comments=True,
                        remove_empty_shapes=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "some_empty_not_remove.shex",
                                                      str_target=str_result))



import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "instantiation_prop" + pth.sep

class TestInstantiationProperty(unittest.TestCase):

    def test_explicit_rdf_type(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        instantiation_property="http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_implicit_rdf_type(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        # instantiation_property="http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_explicit_ex_a(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_BASE_DIR + "G1_ex_a.ttl",
                        instantiation_property="http://example.org/a",
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "G1_ex_a.shex",
                                                      str_target=str_result))

    def test_explicit_ex_a_prefixed(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_BASE_DIR + "G1_ex_a.ttl",
                        instantiation_property="ex:a",
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "G1_ex_a.shex",
                                                      str_target=str_result))

    def test_explicit_ex_a_rdf_type_mixed(self):
        # G1_ex_a_some_rdftype
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=_BASE_DIR + "G1_ex_a_some_rdftype.ttl",
                        instantiation_property="http://example.org/a",
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "G1_ex_a_some_rdftype.shex",
                                                      str_target=str_result))



POSITIVE_CLOSURE = "+"
KLEENE_CLOSURE = "*"
OPT_CARDINALITY = "?"

class Statement(object):

    def __init__(self, st_property, st_type, cardinality, n_occurences,
                 probability, comments=None, serializer_object=None, is_inverse=False):
        self._st_property = st_property
        self._st_type = st_type
        self._cardinality = cardinality
        self._n_occurences = n_occurences
        self._probability = probability
        self._serializer_object = serializer_object
        self._comments = [] if comments is None else comments
        self._is_inverse = is_inverse

    def get_tuples_to_serialize_line_indent_level(self, is_last_statement_of_shape, namespaces_dict):
        return self._serializer_object.\
            serialize_statement_with_indent_level(a_statement=self,
                                                  is_last_statement_of_shape= is_last_statement_of_shape,
                                                  namespaces_dict=namespaces_dict)

    def probability_representation(self):
        return self._serializer_object.probability_representation(self)

    def cardinality_representation(self):
        return self._serializer_object.cardinality_representation(self)

    def comment_representation(self, namespaces_dict):
        return self._serializer_object.turn_statement_into_comment(self, namespaces_dict=namespaces_dict)

    def add_comment(self, comment, insert_first=False):
        if not insert_first:
            self._comments.append(comment)
        else:
            self._comments.insert(0, comment)

    def remove_comments(self):
        self._comments = []


    @property
    def st_property(self):
        return self._st_property

    @property
    def st_type(self):
        return self._st_type

    @property
    def cardinality(self):
        return self._cardinality

    @cardinality.setter
    def cardinality(self, value):
        self._cardinality = value

    @property
    def probability(self):
        return self._probability

    @property
    def n_occurences(self):
        return self._n_occurences

    @probability.setter
    def probability(self, value):
        self._probability = value

    @property
    def comments(self):
        return self._comments

    @property
    def serializer_object(self):
        return self._serializer_object

    @serializer_object.setter
    def serializer_object(self, value):
        self._serializer_object = value

    @property
    def is_inverse(self):
        return self._is_inverse

    @is_inverse.setter
    def is_inverse(self, value):
        self._is_inverse = value



import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, G1_JSON_LD, G1_NT, G1_TSVO_SPO, G1_XML, G1_N3, \
    default_namespaces, G1_TTL_WITH_BASE, G1_TTL_WITH_USELESS_BNODE, G1_TTL_WITH_ABSOLUTES, \
    G1_TTL_WITH_SCAPED_QUOTES
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import NT, TSV_SPO, RDF_XML, JSON_LD, N3, TURTLE, TURTLE_ITER

_BASE_DIR = BASE_FILES + "general" + pth.sep

class TestInputFormat(unittest.TestCase):

    def test_ttl(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))


    def test_nt(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_NT,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=NT,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))


    def test_n3(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_N3,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=N3,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_tsv_spo(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_TSVO_SPO,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TSV_SPO,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_xml(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_XML,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=RDF_XML,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))


    def test_json_ld(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_JSON_LD,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=JSON_LD,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_ttl_iter(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE_ITER,
                        disable_comments=True,
                        infer_numeric_types_for_untyped_literals=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_ttl_iter_with_base(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_TTL_WITH_BASE,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE_ITER,
                        disable_comments=True,
                        infer_numeric_types_for_untyped_literals=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))


    def test_ttl_iter_with_bnode(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_TTL_WITH_USELESS_BNODE,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE_ITER,
                        disable_comments=True,
                        infer_numeric_types_for_untyped_literals=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_ttl_iter_with_absolute_IRIs(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_TTL_WITH_ABSOLUTES,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE_ITER,
                        disable_comments=True,
                        infer_numeric_types_for_untyped_literals=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_ttl_iter_with_scaped_quotes(self):
        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        graph_file_input=G1_TTL_WITH_SCAPED_QUOTES,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE_ITER,
                        disable_comments=True,
                        infer_numeric_types_for_untyped_literals=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))




import gzip

class GzFileLineReader(object):

    def __init__(self, gz_file):
        self._gz_file = gz_file

    def read_lines(self):
        with gzip.open(self._gz_file, "r") as in_stream:
            for a_line in in_stream:
                yield a_line.decode("utf-8")



import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES
from test.t_utils import graph_comparison_file_vs_str
import os.path as pth
from shexer.consts import SHACL_TURTLE

_BASE_DIR = BASE_FILES + "https" + pth.sep


class TestHttps(unittest.TestCase):


    def test_some_https_uris(self):
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "https.ttl",
                        input_format="turtle",
                        disable_comments=True,
                        )
        str_result = shaper.shex_graph(string_output=True,
                                       output_format=SHACL_TURTLE)

        self.assertTrue(graph_comparison_file_vs_str(file_path=_BASE_DIR + "https_shacl.ttl",
                                                     str_target=str_result))




from shexer.core.instances.pconsts import _S, _O

class BaseStrategyMode(object):

    def __init__(self, annotator_ref):
        self._annotator_ref = annotator_ref
        self._instantiation_property = self._annotator_ref._instantiation_property
        self._instances_dict = self._annotator_ref._instances_dict
        self._instance_tracker = self._annotator_ref._instance_tracker

    def is_relevant_triple(self, a_triple):
        raise NotImplementedError()

    def annotate_triple(self, a_triple):
        raise NotImplementedError()

    def annotate_class(self, a_triple):
        self._instances_dict[a_triple[_S].iri].append(a_triple[_O].iri)

    def annotation_post_parsing(self):
        pass  # By default, do nothing.



import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE, SHAPES_DEFAULT_NAMESPACE

_BASE_DIR = BASE_FILES + "sort" + pth.sep # We just need something with another instantiation property


class TestSort(unittest.TestCase):

    def test_outgoing_links(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "g_sort.ttl",
            namespaces_dict=default_namespaces(),
            target_classes=["foaf:Person"],
            input_format=TURTLE,
            disable_comments=True,
            shapes_namespace=SHAPES_DEFAULT_NAMESPACE)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g_sort.shex",
                                                      str_target=str_result,
                                                      check_order=True))

    def test_incoming_links(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "g_sort_incoming.ttl",
            namespaces_dict=default_namespaces(),
            target_classes=["foaf:Person"],
            input_format=TURTLE,
            disable_comments=True,
            shapes_namespace=SHAPES_DEFAULT_NAMESPACE,
            inverse_paths=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g_sort_incoming.shex",
                                                      str_target=str_result,
                                                      check_order=True))

    def test_mixed_links(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "g_sort_mixed.ttl",
            namespaces_dict=default_namespaces(),
            target_classes=["foaf:Person"],
            input_format=TURTLE,
            disable_comments=True,
            shapes_namespace=SHAPES_DEFAULT_NAMESPACE,
            inverse_paths=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g_sort_mixed.shex",
                                                      str_target=str_result,
                                                      check_order=True))


import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces
from test.t_utils import file_vs_str_tunned_comparison, file_vs_str_shex_exact_comparison_excluding_prefixes
from shexer.consts import ALL_EXAMPLES, SHAPE_EXAMPLES, CONSTRAINT_EXAMPLES, ABSOLUTE_INSTANCES
import os.path as pth

from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "example_features" + pth.sep


class TestExamplesMode(unittest.TestCase):

    def test_examples_disabled(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "single_class_deterministic_order.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            examples_mode=None)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "determ_no_examples_no_inverse.shex",
                                                      str_target=str_result))

    def test_constraint_examples_no_inverse(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "single_class_deterministic_order.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=False,
            examples_mode=CONSTRAINT_EXAMPLES)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_shex_exact_comparison_excluding_prefixes(file_path=_BASE_DIR + "determ_constraint_examples_no_inverse.shex",
                                                                             str_target=str_result))

    def test_constraint_examples_with_inverse(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "single_class_deterministic_order.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=False,
            examples_mode=CONSTRAINT_EXAMPLES,
            inverse_paths=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_shex_exact_comparison_excluding_prefixes(file_path=_BASE_DIR + "determ_constraint_examples_with_inverse.shex",
                                                                             str_target=str_result))

    def test_shape_examples_no_instance_stats(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "single_instance_single_prop.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=False,
            examples_mode=SHAPE_EXAMPLES,
            inverse_paths=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_shex_exact_comparison_excluding_prefixes(
            file_path=_BASE_DIR + "determ_single_instance_single_prop_only_shape_examples_no_count.shex",
            str_target=str_result))

    def test_shape_examples_with_instance_stats(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "single_instance_single_prop.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=False,
            examples_mode=SHAPE_EXAMPLES,
            inverse_paths=False,
            instances_report_mode=ABSOLUTE_INSTANCES)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_shex_exact_comparison_excluding_prefixes(
            file_path=_BASE_DIR + "determ_single_instance_single_prop_only_shape_examples_with_count.shex",
            str_target=str_result))

    def test_all_examples_with_instance_stats(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "single_instance_2_prop.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=False,
            examples_mode=ALL_EXAMPLES,
            inverse_paths=False,
            instances_report_mode=ABSOLUTE_INSTANCES)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue('// rdfs:comment "22" ;' in str_result)
        self.assertTrue(":Person   # 1 instance." in str_result)
        self.assertTrue("} // rdfs:comment ex:Jimmy" in str_result)

    def test_all_examples_no_instance_stats(self):
        shaper = Shaper(
            graph_file_input=_BASE_DIR + "single_instance_2_prop.ttl",
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=False,
            examples_mode=ALL_EXAMPLES,
            inverse_paths=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue('// rdfs:comment "22" ;' in str_result)
        self.assertTrue("} // rdfs:comment ex:Jimmy" in str_result)
        self.assertFalse(":Person   # 1 instance." in str_result)






from shexer.io.graph.yielder.multi_nt_triples_yielder import MultiNtTriplesYielder
from shexer.io.graph.yielder.nt_triples_yielder import NtTriplesYielder
from shexer.io.graph.yielder.tsv_nt_triples_yielder import TsvNtTriplesYielder
from shexer.io.graph.yielder.multi_tsv_nt_triples_yielder import MultiTsvNtTriplesYielder
from shexer.io.graph.yielder.rdflib_triple_yielder import RdflibParserTripleYielder, RdflibTripleYielder
from shexer.io.graph.yielder.multi_rdflib_triple_yielder import MultiRdfLibTripleYielder
from shexer.io.graph.yielder.remote.sgraph_from_selectors_triple_yielder import SgraphFromSelectorsTripleYielder
from shexer.io.graph.yielder.filter.filter_namespaces_triple_yielder import FilterNamespacesTriplesYielder
from shexer.io.graph.yielder.big_ttl_triples_yielder import BigTtlTriplesYielder
from shexer.io.graph.yielder.multi_big_ttl_files_triple_yielder import MultiBigTtlTriplesYielder
from shexer.io.graph.yielder.multi_zip_triples_yielder import MultiZipTriplesYielder
from shexer.utils.factories.shape_map_parser_factory import get_shape_map_parser
from shexer.model.graph.endpoint_sgraph import EndpointSGraph
from shexer.utils.translators.list_of_classes_to_shape_map import ListOfClassesToShapeMap
from shexer.utils.target_elements import tune_target_classes_if_needed
from shexer.utils.dict import reverse_keys_and_values
from shexer.utils.compression import list_of_zip_internal_files
from zipfile import ZipFile

from shexer.consts import NT, TSV_SPO, N3, TURTLE, RDF_XML, FIXED_SHAPE_MAP, JSON_LD, TURTLE_ITER, ZIP


def produce_shape_map_according_to_input(sm_format, sgraph, namespaces_prefix_dict, target_classes,
                                         file_target_classes, shape_map_file, shape_map_raw,
                                         instantiation_property, shape_map_already_built=None,
                                         limit_remote_instances=-1, all_classes_mode=False):
    if shape_map_already_built is not None:
        return shape_map_already_built
    prefix_namespaces_dict = reverse_keys_and_values(namespaces_prefix_dict)
    if shape_map_raw is not None or shape_map_file is not None:
        shape_map_parser = get_shape_map_parser(format=sm_format,
                                                sgraph=sgraph,
                                                namespaces_prefix_dict=namespaces_prefix_dict)

        return shape_map_parser.parse_shape_map(source_file=shape_map_file,
                                                raw_content=shape_map_raw)
    else:
        translator = ListOfClassesToShapeMap(sgraph=sgraph,
                                             prefix_namespaces_dict=prefix_namespaces_dict)
        if all_classes_mode:
            return translator.str_class_list_to_shape_map_sparql_selectors(str_list=[a_class for
                                                                                     a_class in
                                                                                     sgraph.yield_classes_with_instances()],
                                                                           instantiation_property=instantiation_property,
                                                                           limit_remote_instances=limit_remote_instances)
        else:
            target_classes = tune_target_classes_if_needed(list_target_classes=target_classes,
                                                           prefix_namespaces_dict=prefix_namespaces_dict) \
                if target_classes is not None \
                else read_target_classes_from_file(file_target_classes=file_target_classes,
                                                   prefix_namespaces_dict=prefix_namespaces_dict)

            return translator.str_class_list_to_shape_map_sparql_selectors(str_list=target_classes,
                                                                           instantiation_property=instantiation_property,
                                                                           limit_remote_instances=limit_remote_instances)


def get_triple_yielder(source_file=None, list_of_source_files=None, input_format=NT, namespaces_to_ignore=None,
                       allow_untyped_numbers=False, raw_graph=None, namespaces_dict=None, url_input=None,
                       list_of_url_input=None, rdflib_graph=None, shape_map_file=None, shape_map_raw=None,
                       shape_map_format=FIXED_SHAPE_MAP,
                       track_classes_for_entities_at_last_depth_level=True, depth_for_building_subgraph=1,
                       url_endpoint=None, instantiation_property=None, strict_syntax_with_corners=False,
                       target_classes=None, file_target_classes=None, built_remote_graph=None,
                       built_shape_map=None, limit_remote_instances=-1, inverse_paths=False, all_classes_mode=False,
                       compression_mode=None, disable_endpoint_cache=False):
    zip_base_archives = _get_base_zip_archive_if_needed(source_file, list_of_source_files, compression_mode)
    result = None
    if url_endpoint is not None:
        result = _yielder_for_url_endpoint(built_remote_graph=built_remote_graph,
                                           url_endpoint=url_endpoint,
                                           built_shape_map=built_shape_map,
                                           shape_map_format=shape_map_format,
                                           namespaces_dict=namespaces_dict,
                                           target_classes=target_classes,
                                           file_target_classes=file_target_classes,
                                           shape_map_file=shape_map_file,
                                           shape_map_raw=shape_map_raw,
                                           instantiation_property=instantiation_property,
                                           limit_remote_instances=limit_remote_instances,
                                           all_classes_mode=all_classes_mode,
                                           depth_for_building_subgraph=depth_for_building_subgraph,
                                           track_classes_for_entities_at_last_depth_level=track_classes_for_entities_at_last_depth_level,
                                           strict_syntax_with_corners=strict_syntax_with_corners,
                                           allow_untyped_numbers=allow_untyped_numbers,
                                           inverse_paths=inverse_paths,
                                           disable_endpoint_cache=disable_endpoint_cache)

    elif url_input is not None or list_of_url_input is not None:  # Always use rdflib to parse remote graphs
        result = _yielder_for_url_input(url_input=url_input,
                                        allow_untyped_numbers=allow_untyped_numbers,
                                        raw_graph=raw_graph,
                                        input_format=input_format,
                                        namespaces_dict=namespaces_dict,
                                        list_of_url_input=list_of_url_input)
    elif rdflib_graph is not None:
        result = RdflibTripleYielder(rdflib_graph=rdflib_graph,
                                     namespaces_dict=namespaces_dict)
    elif input_format == NT:
        result = _yielder_for_nt(source_file=source_file,
                                 raw_graph=raw_graph,
                                 allow_untyped_numbers=allow_untyped_numbers,
                                 list_of_source_files=list_of_source_files,
                                 compression_mode=compression_mode,
                                 zip_base_archives=zip_base_archives)
    elif input_format == TSV_SPO:
        result = _yielder_for_tsv_spo(source_file=source_file,
                                      allow_untyped_numbers=allow_untyped_numbers,
                                      raw_graph=raw_graph,
                                      list_of_files=list_of_source_files,
                                      compression_mode=compression_mode,
                                      zip_base_archives=zip_base_archives)
    elif input_format == TURTLE_ITER:
        result = _yielder_for_turtle_iter(source_file=source_file,
                                          allow_untyped_numbers=allow_untyped_numbers,
                                          raw_graph=raw_graph,
                                          list_of_files=list_of_source_files,
                                          compression_mode=compression_mode,
                                          zip_base_archives=zip_base_archives)
    elif input_format in [N3, RDF_XML, JSON_LD, TURTLE]:
        result = _yielder_for_rdflib_parser(source_file=source_file,
                                            allow_untyped_numbers=allow_untyped_numbers,
                                            raw_graph=raw_graph,
                                            input_format=input_format,
                                            namespaces_dict=namespaces_dict,
                                            list_of_source_files=list_of_source_files,
                                            compression_mode=compression_mode,
                                            zip_base_archives=zip_base_archives)
    else:
        raise ValueError("Not supported format: " + input_format)

    if namespaces_to_ignore is None:
        return result
    else:
        return FilterNamespacesTriplesYielder(actual_triple_yielder=result,
                                              namespaces_to_ignore=namespaces_to_ignore)


def _yielder_for_compressed_inputs(base_yielders):
    if len(base_yielders) == 1:
        result = base_yielders[0]
        return base_yielders[0]
    return MultiZipTriplesYielder(multiyielders=base_yielders)


def _yielder_for_rdflib_parser(source_file, allow_untyped_numbers, raw_graph,
                               input_format, namespaces_dict, list_of_source_files,
                               compression_mode, zip_base_archives):
    if zip_base_archives is not None:
        # return MultiRdfLibTripleYielder(list_of_files=list_of_zip_internal_files(zip_base_archive),
        #                                 allow_untyped_numbers=allow_untyped_numbers,
        #                                 input_format=input_format,
        #                                 namespaces_dict=namespaces_dict,
        #                                 compression_mode=compression_mode,
        #                                 zip_archive_file=zip_base_archives)
        return _yielder_for_compressed_inputs(
            base_yielders=[MultiRdfLibTripleYielder(list_of_files=list_of_zip_internal_files(a_zip_file),
                                                    allow_untyped_numbers=allow_untyped_numbers,
                                                    input_format=input_format,
                                                    namespaces_dict=namespaces_dict,
                                                    compression_mode=compression_mode,
                                                    zip_archive_file=a_zip_file) for a_zip_file in
                           zip_base_archives])

    elif source_file is not None or raw_graph is not None:
        return RdflibParserTripleYielder(source=source_file,
                                         allow_untyped_numbers=allow_untyped_numbers,
                                         raw_graph=raw_graph,
                                         input_format=input_format,
                                         namespaces_dict=namespaces_dict,
                                         compression_mode=compression_mode)

    else:
        return MultiRdfLibTripleYielder(list_of_files=list_of_source_files,
                                        allow_untyped_numbers=allow_untyped_numbers,
                                        input_format=input_format,
                                        namespaces_dict=namespaces_dict,
                                        compression_mode=compression_mode)


def _yielder_for_turtle_iter(source_file, raw_graph, allow_untyped_numbers, list_of_files,
                             compression_mode, zip_base_archives):
    if zip_base_archives is not None:
        # return MultiBigTtlTriplesYielder(list_of_files=list_of_zip_internal_files(zip_base_archive),
        #                                  compression_mode=compression_mode,
        #                                  allow_untyped_numbers=allow_untyped_numbers,
        #                                  zip_base_archives=zip_base_archives)
        return _yielder_for_compressed_inputs(
            [MultiBigTtlTriplesYielder(list_of_files=list_of_zip_internal_files(a_zip_file),
                                       compression_mode=compression_mode,
                                       allow_untyped_numbers=allow_untyped_numbers,
                                       zip_base_archive=a_zip_file) for a_zip_file in zip_base_archives])
    elif source_file is not None or raw_graph is not None:
        return BigTtlTriplesYielder(source_file=source_file,
                                    allow_untyped_numbers=allow_untyped_numbers,
                                    raw_graph=raw_graph,
                                    compression_mode=compression_mode)
    else:
        return MultiBigTtlTriplesYielder(list_of_files=list_of_files,
                                         allow_untyped_numbers=allow_untyped_numbers,
                                         compression_mode=compression_mode)


def _yielder_for_tsv_spo(source_file, raw_graph, allow_untyped_numbers, list_of_files,
                         compression_mode, zip_base_archives):
    if zip_base_archives is not None:
        # return MultiTsvNtTriplesYielder(list_of_files=list_of_zip_internal_files(zip_base_archive),
        #                                 compression_mode=compression_mode,
        #                                 allow_untyped_numbers=allow_untyped_numbers,
        #                                 zip_base_archives=zip_base_archives)
        return _yielder_for_compressed_inputs(
            [MultiTsvNtTriplesYielder(list_of_files=list_of_zip_internal_files(a_zip_file),
                                      compression_mode=compression_mode,
                                      allow_untyped_numbers=allow_untyped_numbers,
                                      zip_base_archive=a_zip_file) for a_zip_file in zip_base_archives])
    elif source_file is not None or raw_graph is not None:
        return TsvNtTriplesYielder(source_file=source_file,
                                   allow_untyped_numbers=allow_untyped_numbers,
                                   raw_graph=raw_graph,
                                   compression_mode=compression_mode)
    else:
        return MultiTsvNtTriplesYielder(list_of_files=list_of_files,
                                        allow_untyped_numbers=allow_untyped_numbers,
                                        compression_mode=compression_mode)


def read_target_classes_from_file(file_target_classes, prefix_namespaces_dict):
    result = []
    with open(file_target_classes, "r") as in_stream:
        for a_line in in_stream:
            candidate = a_line.strip()
            if candidate != "":
                result.append(candidate)
    return tune_target_classes_if_needed(list_target_classes=result,
                                         prefix_namespaces_dict=prefix_namespaces_dict)


def _yielder_for_url_endpoint(built_remote_graph, url_endpoint, built_shape_map, shape_map_format,
                              namespaces_dict, target_classes, file_target_classes, shape_map_file,
                              shape_map_raw, instantiation_property, limit_remote_instances, all_classes_mode,
                              depth_for_building_subgraph, track_classes_for_entities_at_last_depth_level,
                              strict_syntax_with_corners, allow_untyped_numbers, inverse_paths, disable_endpoint_cache):
    sgrpah = built_remote_graph if built_remote_graph is not None else EndpointSGraph(endpoint_url=url_endpoint,
                                                                                      store_locally=not disable_endpoint_cache)

    shape_map = built_shape_map
    if built_shape_map is None:
        shape_map = produce_shape_map_according_to_input(sm_format=shape_map_format,
                                                         sgraph=sgrpah,
                                                         namespaces_prefix_dict=namespaces_dict,
                                                         target_classes=target_classes,
                                                         file_target_classes=file_target_classes,
                                                         shape_map_file=shape_map_file,
                                                         shape_map_raw=shape_map_raw,
                                                         instantiation_property=instantiation_property,
                                                         limit_remote_instances=limit_remote_instances,
                                                         all_classes_mode=all_classes_mode)
    return SgraphFromSelectorsTripleYielder(shape_map=shape_map,
                                            depth=depth_for_building_subgraph,
                                            classes_at_last_level=track_classes_for_entities_at_last_depth_level,
                                            instantiation_property=instantiation_property,
                                            strict_syntax_with_corners=strict_syntax_with_corners,
                                            allow_untyped_numbers=allow_untyped_numbers,
                                            inverse_paths=inverse_paths)


def _yielder_for_url_input(url_input, allow_untyped_numbers, raw_graph,
                           input_format, namespaces_dict, list_of_url_input):
    if url_input:
        return RdflibParserTripleYielder(source=url_input,
                                         allow_untyped_numbers=allow_untyped_numbers,
                                         raw_graph=raw_graph,
                                         input_format=input_format,
                                         namespaces_dict=namespaces_dict)
    else:  # elif list_of_url_input:
        return MultiRdfLibTripleYielder(list_of_files=list_of_url_input,
                                        allow_untyped_numbers=allow_untyped_numbers,
                                        input_format=input_format,
                                        namespaces_dict=namespaces_dict)


def _yielder_for_nt(source_file, raw_graph, allow_untyped_numbers,
                    list_of_source_files, compression_mode,
                    zip_base_archives):
    if (source_file is not None or raw_graph is not None) and zip_base_archives is None:
        return NtTriplesYielder(source_file=source_file,
                                allow_untyped_numbers=allow_untyped_numbers,
                                raw_graph=raw_graph,
                                compression_mode=compression_mode)
    elif zip_base_archives is not None:
        # return MultiNtTriplesYielder(list_of_files=list_of_zip_internal_files(zip_base_archive),
        #                              allow_untyped_numbers=allow_untyped_numbers,
        #                              compression_mode=compression_mode,
        #                              zip_base_archives=zip_base_archives)
        return _yielder_for_compressed_inputs(
            [MultiNtTriplesYielder(list_of_files=list_of_zip_internal_files(a_zip_file),
                                   allow_untyped_numbers=allow_untyped_numbers,
                                   compression_mode=compression_mode,
                                   zip_base_archive=a_zip_file) for a_zip_file in zip_base_archives])

    else:
        return MultiNtTriplesYielder(list_of_files=list_of_source_files,
                                     allow_untyped_numbers=allow_untyped_numbers,
                                     compression_mode=compression_mode)


def _get_base_zip_archive_if_needed(source_file, list_of_source_files, compression_mode):
    if compression_mode != ZIP:
        return None
    if source_file is not None:
        return [ZipFile(source_file, 'r')]
    result = []
    for a_source_file in list_of_source_files:
        result.append(ZipFile(a_source_file, 'r'))
    return result



import unittest
from shexer.shaper import Shaper
from test.const import BASE_FILES, default_namespaces
from test.t_utils import file_vs_str_exact_comparison
import os.path as pth
from shexer.consts import TURTLE, MIXED_INSTANCES, ABSOLUTE_INSTANCES, RATIO_INSTANCES

_BASE_DIR = BASE_FILES + "freq_reports" + pth.sep

class TestGraphFileInput(unittest.TestCase):


    def test_comments_disabled(self):
        """
        Even if a instances report mode is configured, with comments disabled there shouldnt be any
        other content than the pure shape

        :return:
        """
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "g_person_infinite_frequencies.ttl",
                        namespaces_dict=default_namespaces(),
                        input_format=TURTLE,
                        disable_comments=True,
                        instances_report_mode=ABSOLUTE_INSTANCES)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_exact_comparison(target_str=str_result,
                                                     file_path=_BASE_DIR + "g_person_comments_disabled.shex"))

    def test_ratio_mode(self):
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "g_person_infinite_frequencies.ttl",
                        namespaces_dict=default_namespaces(),
                        input_format=TURTLE,
                        disable_comments=False,
                        instances_report_mode=RATIO_INSTANCES)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_exact_comparison(target_str=str_result,
                                                     file_path=_BASE_DIR + "g_person_every_decimal.shex"))

    def test_absolute_mode(self):
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "g_person_infinite_frequencies.ttl",
                        namespaces_dict=default_namespaces(),
                        input_format=TURTLE,
                        disable_comments=False,
                        instances_report_mode=ABSOLUTE_INSTANCES)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_exact_comparison(target_str=str_result,
                                                     file_path=_BASE_DIR + "g_person_absolute_instances.shex"))

    def test_absolute_mode_unbound_dec(self):
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "g_person_infinite_frequencies.ttl",
                        namespaces_dict=default_namespaces(),
                        input_format=TURTLE,
                        disable_comments=False,
                        instances_report_mode=MIXED_INSTANCES)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_exact_comparison(target_str=str_result,
                                                     file_path=_BASE_DIR + "g_person_mixed_unbound_dec.shex"))

    def test_absolute_mode_2_dec(self):
        shaper = Shaper(all_classes_mode=True,
                        graph_file_input=_BASE_DIR + "g_person_infinite_frequencies.ttl",
                        namespaces_dict=default_namespaces(),
                        input_format=TURTLE,
                        disable_comments=False,
                        instances_report_mode=MIXED_INSTANCES,
                        decimals=2)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_exact_comparison(target_str=str_result,
                                                     file_path=_BASE_DIR + "g_person_mixed_2_dec.shex"))



from shexer.core.instances.abstract_instance_tracker import AbstractInstanceTracker
from shexer.utils.log import log_msg


class ShapeMapInstanceTracker(AbstractInstanceTracker):

    def __init__(self, shape_map):
        self._shape_map = shape_map
        self._instances_dict = {}

    def track_instances(self, verbose=False):
        log_msg(verbose=verbose,
                msg="Starting instance tracker...")
        for an_item in self._shape_map.yield_items():
            self._solve_targets_of_an_item(an_item)
        log_msg(verbose=verbose,
                msg="Instance tracker finished. {} instances located".format(len(self._instances_dict)))
        return self._instances_dict

    def _solve_targets_of_an_item(self, an_item):
        for a_node in an_item.node_selector.get_target_nodes():
            if a_node not in self._instances_dict:
                self._instances_dict[a_node] = []
            self._instances_dict[a_node].append(an_item.shape_label)

    def _specific_disambiguator_prefix(self):
        return "custom_"




from shexer.core.instances.annotators.strategy_mode.base_strategy_mode import BaseStrategyMode
from shexer.core.instances.pconsts import _P, _O

class TargetClassesMode(BaseStrategyMode):

    def __init__(self, annotator_ref):
        super().__init__(annotator_ref)
        self._target_classes = self._annotator_ref._target_classes

    def is_relevant_triple(self, a_triple):
        if a_triple[_P] != self._instantiation_property:
            return False
        if a_triple[_O] not in self._target_classes:
            return False
        return True

    def annotate_triple(self, a_triple):
        if self._instance_tracker.is_an_instantiation_prop(a_triple[_P]):
            self._annotator_ref.add_instance_to_instances_dict(a_triple)
            self.annotate_class(a_triple)




from shexer.io.graph.yielder.nt_triples_yielder import NtTriplesYielder
from shexer.io.graph.yielder.multifile_base_triples_yielder import MultifileBaseTripleYielder


class MultiNtTriplesYielder(MultifileBaseTripleYielder):

    def __init__(self, list_of_files,
                 allow_untyped_numbers=False,
                 compression_mode=None,
                 zip_base_archive=None):
        super(MultiNtTriplesYielder, self).__init__(list_of_files=list_of_files,
                                                    allow_untyped_numbers=allow_untyped_numbers,
                                                    compression_mode=compression_mode,
                                                    zip_base_archive=zip_base_archive)

    def _constructor_file_yielder(self, a_source_file, parse_namespaces=False):
        return NtTriplesYielder(source_file=a_source_file,
                                allow_untyped_numbers=self._allow_untyped_numbers,
                                compression_mode=self._compression_mode,
                                zip_base_archive=self._zip_base_archive)



from shexer.core.profiling.class_profiler import RDF_TYPE_STR
from shexer.model.shape import STARTING_CHAR_FOR_SHAPE_NAME
from rdflib import Graph, Namespace, URIRef, RDF, BNode, XSD, Literal
from shexer.model.statement import POSITIVE_CLOSURE, KLEENE_CLOSURE, OPT_CARDINALITY
from shexer.utils.uri import XSD_NAMESPACE, LANG_STRING_TYPE
from shexer.model.const_elem_types import IRI_ELEM_TYPE, LITERAL_ELEM_TYPE, DOT_ELEM_TYPE, BNODE_ELEM_TYPE
from shexer.io.wikidata import wikidata_annotation
from wlighter import TURTLE_FORMAT

_EXPECTED_SHAPE_BEGINING = STARTING_CHAR_FOR_SHAPE_NAME + "<"
_EXPECTED_SHAPE_ENDING = ">"

_SHACL_NAMESPACE = "http://www.w3.org/ns/shacl#"

_SHACL_PRIORITY_PREFIXES = ["sh", "shacl", "sha"]

_R_SHACL_SHAPE_URI = URIRef(_SHACL_NAMESPACE + "NodeShape")
_R_SHACL_PROPERTY_SHAPE_URI = URIRef(_SHACL_NAMESPACE + "PropertyShape")

_R_SHACL_TARGET_CLASS_PROP = URIRef(_SHACL_NAMESPACE + "targetClass")
_R_SHACL_PATH_PROP = URIRef(_SHACL_NAMESPACE + "path")
_R_SHACL_INVERSE_PATH_PROP = URIRef(_SHACL_NAMESPACE + "inversePath")
_R_SHACL_MIN_COUNT_PROP = URIRef(_SHACL_NAMESPACE + "minCount")
_R_SHACL_MAX_COUNT_PROP = URIRef(_SHACL_NAMESPACE + "maxCount")

_R_SHACL_PROPERTY_PROP = URIRef(_SHACL_NAMESPACE + "property")

_R_SHACL_DATATYPE_PROP = URIRef(_SHACL_NAMESPACE + "dataType")
_R_SHACL_NODEKIND_PROP = URIRef(_SHACL_NAMESPACE + "nodeKind")
_R_SHACL_NODE_PROP = URIRef(_SHACL_NAMESPACE + "node")

_R_SHACL_IN_PROP = URIRef(_SHACL_NAMESPACE + "in")

_R_SHACL_PATTERN_PROP = URIRef(_SHACL_NAMESPACE + "pattern")

_R_SHACL_NODEKIND_IRI = URIRef(_SHACL_NAMESPACE + "IRI")
_R_SHACL_NODEKIND_LITERAL = URIRef(_SHACL_NAMESPACE + "Literal")
_R_SHACL_NODEKIND_BNODE = URIRef(_SHACL_NAMESPACE + "BlankNode")
_R_SHACL_NODEKIND_DOT = None

_R_LANG_STRING = URIRef("http://www.w3.org/2000/01/rdf-schema#langString")

_INTEGER = "i"
_STRING = "s"

_MACRO_MAPPING = {IRI_ELEM_TYPE: _R_SHACL_NODEKIND_IRI,
                  LITERAL_ELEM_TYPE: _R_SHACL_NODEKIND_LITERAL,
                  DOT_ELEM_TYPE: _R_SHACL_NODEKIND_BNODE,
                  BNODE_ELEM_TYPE: _R_SHACL_NODEKIND_DOT}


class ShaclSerializer(object):

    def __init__(self, target_file, shapes_list, namespaces_dict=None, string_return=False,
                 instantiation_property_str=RDF_TYPE_STR, wikidata_annotation=False,
                 detect_minimal_iri=False, shape_example_features=None):
        self._target_file = target_file
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
        self._shapes_list = shapes_list
        self._string_return = string_return
        self._instantiation_property_str = instantiation_property_str
        self._wikidata_annotation = wikidata_annotation
        self._detect_minimal_iri = detect_minimal_iri
        self._shape_example_features = shape_example_features

        self._g_shapes = Graph()

        # self._uri_dict = {}

    def serialize_shapes(self):
        self._add_namespaces()
        self._add_shapes()
        return self._produce_output()

    #################### NAMESPACES

    def _add_namespaces(self):
        self._add_param_namespaces()
        self._add_shacl_namespace_if_needed()

    def _add_param_namespaces(self):
        for a_namespace, a_prefix in self._namespaces_dict.items():
            self._add_namespace(prefix=a_prefix,
                                namespace_str=a_namespace)

    def _add_namespace(self, prefix, namespace_str):
        self._g_shapes.bind(prefix=prefix,
                            namespace=Namespace(namespace_str))

    def _add_shacl_namespace_if_needed(self):
        if _SHACL_NAMESPACE in self._namespaces_dict:  # shacl already included
            return
        curr_prefixes = self._namespaces_dict.values()
        for a_prefix in _SHACL_PRIORITY_PREFIXES:  # trying default prefixes
            if a_prefix not in curr_prefixes:
                self._add_shacl_namespace(a_prefix)
                return
        counter = 1  # going for random prefixes, no defs. available
        candidate_pref = _SHACL_PRIORITY_PREFIXES[0] + str(counter)
        while candidate_pref in curr_prefixes:
            counter += 1
            candidate_pref = _SHACL_PRIORITY_PREFIXES[0] + str(counter)
        self._add_shacl_namespace(candidate_pref)

    def _add_shacl_namespace(self, shacl_prefix):
        self._add_namespace(prefix=shacl_prefix,
                            namespace_str=_SHACL_NAMESPACE)
        self._namespaces_dict[_SHACL_NAMESPACE] = shacl_prefix

    #################### SHAPES

    def _add_shapes(self):
        for a_shape in self._shapes_list:
            self._add_shape(a_shape)

    def _add_shape(self, shape):
        r_shape_uri = self._generate_shape_uri(shape_name=shape.name)
        self._add_shape_uri(r_shape_uri=r_shape_uri)
        self._add_target_class(r_shape_uri=r_shape_uri,
                               shape=shape)
        if self._detect_minimal_iri:
            self._add_min_iri(r_shape_uri=r_shape_uri,
                              shape=shape)
        self._add_shape_constraints(shape=shape,
                                    r_shape_uri=r_shape_uri)


    def _add_target_class(self, shape, r_shape_uri):
        if shape.class_uri is not None:
            self._add_triple(r_shape_uri,
                             _R_SHACL_TARGET_CLASS_PROP,
                             URIRef(shape.class_uri))  # TODO check if this is always an abs. URI, not sure

    def _add_min_iri (self, shape, r_shape_uri):
        # if shape.iri_pattern is not None:
        if self._shape_example_features.shape_min_iri(shape_id=shape.class_uri) is not None:
            self._add_triple(r_shape_uri,
                             _R_SHACL_PATTERN_PROP,
                             self._literal_iri_pattern(shape))

    def _literal_iri_pattern(self, shape):
        return Literal("^{}".format(self._shape_example_features.shape_min_iri(shape_id=shape.class_uri)))

    def _add_shape_constraints(self, shape, r_shape_uri):
        for a_statement in shape.yield_statements():
            self._add_constraint(statement=a_statement,
                                 r_shape_uri=r_shape_uri)

    def _is_instantiation_property(self, str_property):
        return str_property == self._instantiation_property_str

    def _add_constraint(self, statement, r_shape_uri):
        if self._is_instantiation_property(statement.st_property):
            self._add_instantiation_constraint(statement=statement,
                                               r_shape_uri=r_shape_uri)
        else:
            self._add_regular_constraint(statement=statement,
                                         r_shape_uri=r_shape_uri)

    def _add_exactly_one_cardinality(self, r_constraint_node):
        self._add_min_occurs(r_constraint_node=r_constraint_node,
                             min_occurs=1)
        self._add_max_occurs(r_constraint_node=r_constraint_node,
                             max_occurs=1)

    def _add_in_instance(self, r_constraint_node, statement):
        target_node = self._generate_r_uri_for_str_uri(statement.st_type)
        list_seed_node = self._generate_bnode()
        self._add_triple(r_constraint_node, _R_SHACL_IN_PROP, list_seed_node)
        self._add_triple(list_seed_node, RDF.first, target_node)
        self._add_triple(list_seed_node, RDF.rest, RDF.nil)

    def _add_instantiation_constraint(self, statement, r_shape_uri):
        r_constraint_node = self._generate_bnode()
        self._add_bnode_property(r_shape_uri=r_shape_uri,
                                 r_constraint_node=r_constraint_node)
        self._add_direct_path(statement=statement,
                              r_constraint_node=r_constraint_node)
        self._add_exactly_one_cardinality(r_constraint_node=r_constraint_node)
        self._add_in_instance(statement=statement,
                              r_constraint_node=r_constraint_node)

    def _add_regular_constraint(self, statement, r_shape_uri):
        r_constraint_node = self._generate_bnode()
        self._add_bnode_property(r_shape_uri=r_shape_uri,
                                 r_constraint_node=r_constraint_node)
        self._add_node_type(statement=statement,
                            r_constraint_node=r_constraint_node)
        self._add_cardinality(statement=statement,
                              r_constraint_node=r_constraint_node)
        self._add_path(statement=statement,
                       r_constraint_node=r_constraint_node)

    def _add_path(self, statement, r_constraint_node):
        if not statement.is_inverse:
            self._add_direct_path(statement=statement,
                                  r_constraint_node=r_constraint_node)
        else:
            self._add_inverse_path(statement=statement,
                                   r_constraint_node=r_constraint_node)

    def _add_direct_path(self, statement, r_constraint_node):
        r_property_uri = self._generate_r_uri_for_str_uri(statement.st_property)
        self._add_triple(r_constraint_node, _R_SHACL_PATH_PROP, r_property_uri)

    def _add_inverse_path(self, statement, r_constraint_node):
        r_property_uri = self._generate_r_uri_for_str_uri(statement.st_property)
        inverse_path_node = self._generate_bnode()
        self._add_triple(r_constraint_node, _R_SHACL_PROPERTY_PROP, inverse_path_node)
        self._add_triple(inverse_path_node, _R_SHACL_INVERSE_PATH_PROP, r_property_uri)

    def _generate_r_uri_for_str_uri(self, property_str):
        if property_str.startswith("<") and property_str.endswith(">"):
            return URIRef(property_str[1:-1])
        elif property_str.startswith("http://") or property_str.startswith("https://"):
            return URIRef(property_str)
        raise ValueError("Having troubles recognizing this URI", property_str, ". "
                        "Is it well-formed? If you think so, add a GitHub issue. ")

    def _is_a_shape(self, target_type):
        return target_type.startswith(STARTING_CHAR_FOR_SHAPE_NAME)

    def _is_literal(self, target_type):
        return target_type == LANG_STRING_TYPE or target_type.startswith(XSD_NAMESPACE)

    def _is_macro(self, target_type):
        return target_type in _MACRO_MAPPING

    def _add_dataType_literal(self, r_constraint_node, target_type):
        # if target_type == LANG_STRING_TYPE:
        #     type_node = _R_LANG_STRING
        # elif target_type.endswith("integer"):
        #     type_node = XSD.integer
        # elif target_type.endswith("float"):
        #     type_node = XSD.float
        # elif target_type.endswith("string"):
        #     type_node = XSD.string
        # else:
        #     raise ValueError("Unexpected literal type:" + target_type)
        self._add_triple(r_constraint_node,
                         _R_SHACL_DATATYPE_PROP,
                         URIRef(target_type))

    def _add_node_shape(self, r_constraint_node, target_type):
        self._add_triple(r_constraint_node,
                         _R_SHACL_NODE_PROP,
                         self._generate_shape_uri(shape_name=target_type))

    def _add_nodeKind_macro(self, r_constraint_node, target_type):
        type_node = _MACRO_MAPPING[target_type]
        if type_node is not None:
            self._add_triple(r_constraint_node,
                             _R_SHACL_NODEKIND_PROP,
                             type_node)

    def _add_node_type(self, statement, r_constraint_node):
        #  sh:dataType for literal types
        #  sh:nodeKind for IRI or similar macros.
        #  sh:node for a shape
        # if self._is_literal(statement.st_type):
        #     self._add_dataType_literal(r_constraint_node=r_constraint_node,
        #                                target_type=statement.st_type)
        if self._is_macro(statement.st_type):
            self._add_nodeKind_macro(r_constraint_node=r_constraint_node,
                                     target_type=statement.st_type)
        elif self._is_a_shape(statement.st_type):
            self._add_node_shape(r_constraint_node=r_constraint_node,
                                 target_type=statement.st_type)
        else:  # It should be a literal
            self._add_dataType_literal(r_constraint_node=r_constraint_node,
                                       target_type=statement.st_type)
        # else:
        #     raise ValueError("Check here: ")


    def _min_occurs_from_cardinality(self, cardinality):
        if cardinality in [KLEENE_CLOSURE, OPT_CARDINALITY]:
            return None
        elif cardinality == POSITIVE_CLOSURE:
            return 1
        else:
            return cardinality

    def _max_occurs_from_cardinality(self, cardinality):
        if cardinality in [KLEENE_CLOSURE, POSITIVE_CLOSURE]:
            return None
        elif cardinality == OPT_CARDINALITY:
            return 1
        else:
            return cardinality

    def _generate_r_literal(self, value, l_type):
        return Literal(value, datatype=self._map_rdflib_datatype(l_type))

    def _map_rdflib_datatype(self, l_type):
        if l_type == _INTEGER:
            return XSD.integer
        elif l_type == _STRING:
            return XSD.string
        else:
            raise ValueError("Having troubles recognizing this literal type", l_type, ". "
                        "Is it well-formed? If you think so, add a GitHub issue. ")

    def _add_min_occurs(self, r_constraint_node, min_occurs):
        self._add_triple(r_constraint_node,
                         _R_SHACL_MIN_COUNT_PROP,
                         self._generate_r_literal(value=min_occurs,
                                                  l_type=_INTEGER))

    def _add_max_occurs(self, r_constraint_node, max_occurs):
        self._add_triple(r_constraint_node,
                         _R_SHACL_MAX_COUNT_PROP,
                         self._generate_r_literal(value=max_occurs,
                                                  l_type=_INTEGER))

    def _add_cardinality(self, statement, r_constraint_node):
        min_occurs = self._min_occurs_from_cardinality(statement.cardinality)
        max_occurs = self._max_occurs_from_cardinality(statement.cardinality)
        if min_occurs is not None:
            self._add_min_occurs(r_constraint_node=r_constraint_node,
                                 min_occurs=min_occurs)
        if max_occurs is not None:
            self._add_max_occurs(r_constraint_node=r_constraint_node,
                                 max_occurs=max_occurs)

    def _add_bnode_property(self, r_shape_uri, r_constraint_node):
        self._add_triple(r_shape_uri, _R_SHACL_PROPERTY_PROP, r_constraint_node)
        self._add_triple(r_constraint_node, RDF.type, _R_SHACL_PROPERTY_SHAPE_URI)

    def _generate_shape_uri(self, shape_name):
        if shape_name.startswith(_EXPECTED_SHAPE_BEGINING) and shape_name.endswith(_EXPECTED_SHAPE_ENDING):
            return URIRef(shape_name[2:-1])  # Excluding  "@<"  and ">
        raise ValueError("Unknown error, having trouble with a shape label:", shape_name,
                         "Add a GitHub issue to github with your input to have this review and fixed.")

    def _add_shape_uri(self, r_shape_uri):
        self._add_triple(r_shape_uri, RDF.type, _R_SHACL_SHAPE_URI)

    def _add_triple(self, s, p, o):
        self._g_shapes.add((s, p, o))

    @staticmethod
    def _generate_bnode():
        return BNode()

    #################### OUTPUT

    def _produce_output(self):
        if self._wikidata_annotation:
            return self._produce_wikidata_annotation_output()
        # destination = None if self._string_return else self._target_file
        if self._string_return:
            return self._g_shapes.serialize(format="turtle")
        else:
            self._g_shapes.serialize(destination=self._target_file, format="turtle")


    def _produce_wikidata_annotation_output(self):
        result = self._g_shapes.serialize(format="turtle")
        result = wikidata_annotation(raw_input=result,
                                     string_return=self._string_return,
                                     out_file=self._target_file,
                                     format=TURTLE_FORMAT,
                                     rdfs_comments=False)
        if self._string_return:
            return result




import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import check_file_exist, delete_file, file_vs_str_tunned_comparison, number_of_shapes
import os.path as pth

from shexer.consts import TURTLE



_BASE_DIR = BASE_FILES + "uml" + pth.sep
_A_PATH_FOR_IMG = _BASE_DIR + "any_path.png"


class TestUmlGen(unittest.TestCase):

    def test_base_gen(self):
        delete_file(_A_PATH_FOR_IMG)  # pre-condition
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True)
        shaper.shex_graph(to_uml_path=_A_PATH_FOR_IMG)
        try:
            check_file_exist(_A_PATH_FOR_IMG)
        finally:
            delete_file(_A_PATH_FOR_IMG)

    def test_gen_uml_and_classical_output(self):

        delete_file(_A_PATH_FOR_IMG)  # pre-condition
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True)
        str_result = shaper.shex_graph(to_uml_path=_A_PATH_FOR_IMG, string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))
        try:
            check_file_exist(_A_PATH_FOR_IMG)
        finally:
            delete_file(_A_PATH_FOR_IMG)

    def test_wikidata(self):
        shape_map_raw = "SPARQL'select ?p where " \
                        "{ ?p <http://www.wikidata.org/prop/direct/P31> <http://www.wikidata.org/entity/Q14660> } " \
                        "LIMIT 1'@<Flag>"
        shaper = Shaper(shape_map_raw=shape_map_raw,
                        url_endpoint="https://query.wikidata.org/sparql",
                        namespaces_dict=default_namespaces(),
                        instantiation_property="http://www.wikidata.org/prop/direct/P31",
                        disable_comments=True,
                        depth_for_building_subgraph=1,
                        track_classes_for_entities_at_last_depth_level=False,
                        all_classes_mode=False)
        str_result = shaper.shex_graph(string_output=True, to_uml_path=_A_PATH_FOR_IMG)
        self.assertTrue(number_of_shapes(str_result) == 1)
        try:
            check_file_exist(_A_PATH_FOR_IMG)
        finally:
            delete_file(_A_PATH_FOR_IMG)

